import kego.plotting.analysis
import kego.plotting.axes_utils
import kego.plotting.colormesh
import kego.plotting.figures
import kego.plotting.histograms
import kego.plotting.lines
import kego.plotting.timeseries
import kego.plotting.value_counts
from kego.plotting.histograms import plot_histogram, plot_histogram_2d
